from typing import Literal
from Search import Search
import requests
from bs4 import BeautifulSoup
from Validation import Validations


class StatsSearch(Search):
    def __init__(
        self,
        year: str | int,
        month: Literal["Must be a two digit month string eg: '01' for January"],
        gen: str | int,
    ) -> None:
        super().__init__(year, month, gen)
        self.ending = None

    ### NOT USED IN CHAOS BRANCH SEARCHES ###
    def _remove_formatting(self, data: list, isMonotype: bool) -> list[list]:
        # Set the column names first
        outlist = [
            [
                "rank",
                "pokemon",
                "usage_pct",
                "raw_usage",
                "raw_pct",
                "real",
                "real_pct",
                "date",
                "tier",
            ]
        ]
        # Remove the formatting from the webpage
        for line in data:
            # remove all linebreaks,symbols and make lowercase
            line = line.strip("\n")
            line = line.replace("|", ",").replace(" ", "").replace("%", "").lower()
            if line.startswith(","):
                # remove leading and trailing commas
                line = line[1:-1]
                # turns it from a list of strings toa 2-d array
                line = line.split(",")
                line.append(f"{self.year}-{self.month}")
                if not isMonotype:
                    line.append(f"gen{self.gen}{self.tier}")
                elif isMonotype:
                    line.append(f"mono{self.typing}")

                # add it to the list to be returned
                outlist.append(line)
        return outlist

    @staticmethod
    def _vintage_search(year, month, gen, tier) -> str:
        r = requests.get(f"https://www.smogon.com/stats/{year}-{month}/")
        soup = BeautifulSoup(r.text, "html.parser")
        anchors = soup.find_all("a")

        tiers = tuple(set(a.text.strip(".txt").split("-")[0] for a in anchors[6:]))

        try:
            if tiers.index(f"gen{gen}{tier}"):
                return f"gen{gen}{tier}-1500.txt"
        except ValueError:
            try:
                if tiers.index(tier) and tier == (6 | 7):
                    return f"{tier}-1500.txt"
            except ValueError:
                return "Nothing found."

    def search(self) -> list[list]:
        """
        Searches the smogon stats repository and returns a list of lists.
        """
        # send the request to get the data
        res = requests.get(self.base)
        # get the text if there is any (should add a check in here)
        data = res.text
        # send each "unit" of information to a list by splitting on linebreaks
        data = data.split("\n")
        # truncate to get rid of the non-relevant prefixes and last two lines of styling
        data = data[5:-2]
        # remove the formatting
        data = self._remove_formatting(data, isMonotype=self.isMonotype)
        return data


class BaseStatsSearch(StatsSearch):
    """>tier (str): string of tier to be queried."""

    __doc__ = Search.__doc__ + __doc__

    def __init__(
        self,
        year: str | int,
        month: Literal["Must be a two digit month string eg: '01' for Janu…"],
        gen: str | int,
        tier: str,
    ) -> None:
        super().__init__(year, month, gen)
        self.tier = tier.lower()
        self.base += f"{self.year}-{self.month}/"
        self._build_url()

    @property
    def tier(self) -> str:
        return self._tier.lower()

    @tier.setter
    def tier(self, value):
        self._tier = value

    def _build_url(self):
        # Rating of 1500 is defined by Smogon as their target of a normal player

        validation_object = self._create_validation_object()
        validator = Validations(validation_object)
        # check dates for valid tiers
        # since for now I am only looking up newer stats for database, it will be fine
        if validator.validate():
            if Validations.is_modern_format(validation_object):
                ending = f"gen{self.gen}{self.tier}-1500.txt"
                self.ending = ending
                self.base += ending
            else:
                ending = self._vintage_search(
                    self.year, self.month, self.gen, self.tier
                )
                # below can be deleted after validations are done
                if ending.endswith(".txt"):
                    self.ending = ending
                    self.base += ending
                # else:
                #     raise Exception('Invalid information supplied, check your parameters.')
        else:
            print("Something isn't correct.")


class MonotypeStatsSearch(StatsSearch):
    """>typing (str): string of type to be queried."""

    __doc__ = Search.__doc__ + __doc__

    def __init__(
        self,
        year: str | int,
        month: Literal["Must be a two digit month string eg: '01' for Janu…"],
        gen: str | int,
        typing: Literal["Must be string of type, such as 'fighting' or 'psychic'"],
    ) -> None:
        super().__init__(year, month, gen)
        self.typing = typing.lower()
        self.base += f"{year}-{month}/monotype/"
        self._build_url()
        self.isMonotype = True

    @property
    def typing(self) -> str:
        return self._typing.lower()

    @typing.setter
    def typing(self, value):
        self._typing = value

    # I could probably combine this and the other into a singular _build_url() that lives in the parent class, but I think this is simpler
    def _build_url(self):
        # Rating of 1500 is defined by Smogon as their target of a normal player

        validation_object = self._create_validation_object()
        validator = Validations(validation_object)
        print(validation_object)
        # check dates for valid tiers
        # since for now I am only looking up newer stats for database, it will be fine
        if validator.validate(isMonotype=True):
            ending = f"gen{self.gen}monotype-mono{self.typing}-1500.txt"
            self.ending = ending
            self.base += ending
        else:
            print("Something isn't correct.")


if __name__ == "__main__":
    base_search = BaseStatsSearch("2023", "06", "9", "ou")
    base_search_data = base_search.search()
    print(base_search._create_validation_object())
    print(base_search_data[3])

    mono_search = MonotypeStatsSearch("2022", "06", "8", "psychic")
    mono_search_data = mono_search.search()
    print(mono_search._create_validation_object())
    print(mono_search_data[0:5])



from datetime import datetime

types = [
    "normal",
    "fire",
    "water",
    "grass",
    "flying",
    "fighting",
    "poison",
    "electric",
    "ground",
    "rock",
    "psychic",
    "ice",
    "bug",
    "ghost",
    "steel",
    "dragon",
    "dark",
    "fairy",
]


class Validations:
    def __init__(self, validation_object: dict) -> None:
        self.validation_object = validation_object

    def validate(self, isMonotype: bool = False) -> bool:
        """
        Runs the validation checks, if any of them return False, this also returns False.
        """
        if isMonotype:
            if self.validation_object["typing"].lower() not in types:
                print(f"Invalid type {self.validation_object['typing']} is invalid.")
                return False
            if not (self._is_valid_date(self.validation_object)):
                print("Date is invalid.")
                return False
        else:
            if not (self._is_correct_gen(self.validation_object)):
                print("Generation choice is invalid")
                return False
            if not (self._is_valid_date(self.validation_object)):
                print("Date and generation combination is invalid.")
                return False
        return True

    @staticmethod
    def is_modern_format(validation_object: dict) -> bool:
        """Checks if the date submitted uses the old non-standardized labeling or not. Anything before Jul 2017 uses non-standard."""
        if (int(validation_object["date"].year) == 2017) and (
            int(validation_object["date"].month) <= 6
        ):
            return False
        elif int(validation_object["date"].year) < 2017:
            return False
        return True

    @staticmethod
    def _is_correct_gen(validation_object: dict) -> bool:
        """Checks if the generation is valid for the date submitted."""
        cutoff_dates = {
            "7": datetime.strptime("2016-11", "%Y-%m"),
            "8": datetime.strptime("2020-11", "%Y-%m"),
            "9": datetime.strptime("2022-11", "%Y-%m"),
        }
        # something like having an object of dates this is where it makes sense to
        # have dt objects rather than strings
        # if they fail, return false.

        if int(validation_object["gen"]) <= 6:
            return True
        elif (int(validation_object["gen"]) > 6) and (
            validation_object["date"] < cutoff_dates[str(validation_object["gen"])]
        ):
            return False
        else:
            return True

    @staticmethod
    def _is_valid_date(validation_object: dict) -> bool:
        """Makes sure theres no date submitted before Nov 2014, as there is no data available before then."""
        year = int(validation_object["date"].year)
        month = int(validation_object["date"].month)

        if year < 2014:
            print(f"Year {year} is invalid.")
            return False
        elif year == 2014:
            if month != (11 | 12):
                print(f"Year {year} and month {month} is invalid.")
                return False
        else:
            return True



import requests
from typing import Literal
from Search import Search
from UsageStatsLookup import BaseStatsSearch

# TODO: Account for rating when gen is current gen to ONLY be 1695, otherwise be 1630


# Chaos searches are basically individual pokemon lookups, where one can get deeper analysis
# such as movesets, held items, EV spreads, etc
class ChaosSearch(Search):
    def __init__(
        self,
        year: str | int,
        month: Literal["Must be a two digit month string eg: '01' for January"],
        gen: str | int,
        name: str,
    ) -> None:
        self.name = name.lower()
        super().__init__(year, month, gen)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value.lower()

    def search(self) -> object | None:
        res = requests.get(self.base)
        page_object = res.json()["data"]
        page_object = self.lower_keys(page_object)

        try:
            lookup = page_object[self.name]
            return lookup
        except:
            return None

    # We will be doing this a lot so make it a function.
    def lower_keys(self, dictionary: dict) -> dict:
        """
        Returns a dictionary with all keys lower-cased.
        """
        return {k.lower(): v for k, v in dictionary.items()}


class BaseChaosSearch(ChaosSearch):
    def __init__(
        self,
        year: str | int,
        month: Literal["Must be a two digit month string eg: '01' for January"],
        gen: str | int,
        name: str,
        tier: str,
    ) -> None:
        super().__init__(year, month, gen, name)
        self.tier = tier.lower()
        self.base = f"https://www.smogon.com/stats/{self.year}-{self.month}/chaos/"
        self._build_url()

    @property
    def tier(self) -> str:
        return self._tier

    @tier.setter
    def tier(self, value):
        self._tier = value

    def _build_url(self):
        self.base += f"gen{self.gen}{self.tier}-1500.json"


class MonotypeChaosSearch(ChaosSearch):
    def __init__(
        self,
        year: str | int,
        month: Literal["Must be a two digit month string eg: '01' for January"],
        gen: str | int,
        name: str,
        typing: str,
    ) -> None:
        super().__init__(year, month, gen, name)
        self.base = (
            f"https://www.smogon.com/stats/{self.year}-{self.month}/monotype/chaos/"
        )
        self.typing = typing.lower()
        self.isMonotype = True
        self._build_url()

    @property
    def typing(self) -> str:
        return self._typing

    @typing.setter
    def typing(self, value):
        self._typing = value

    def _build_url(self):
        self.base += f"gen{self.gen}monotype-mono{self.typing}-1500.json"


# this is to search stats individually, but its a bit more involved, a WIP
class IndividualStatsSearch(BaseStatsSearch):
    def __init__(
        self,
        year: str | int,
        month: Literal["Must be a two digit month string eg: '01' for Janu…"],
        gen: str | int,
        tier: str,
        name: str,
    ) -> None:
        super().__init__(year, month, gen, tier)
        self.name = name.lower()

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value.lower()

    def find_individual_usage(self, resmatrix: list[list] = None) -> dict | None:
        # resmatrix is the resulting 2d array from the self.search() function
        if not resmatrix:
            resmatrix = self.search()
        for m in resmatrix:
            if m[1] == self.name:
                return dict(zip(resmatrix[0], m))
        print("No result found.")
        return None


if __name__ == "__main__":
    # search = BaseChaosSearch(year="2022", month="07", gen=8, tier="uu", name="SKARMORY")
    # result = search.search()
    # print(f"{search.name}:  {result}\n")

    # monosearch = MonotypeChaosSearch(
    #     year=2022, month="11", gen=9, typing="fairy", name="Gardevoir"
    # )
    # monoresult = monosearch.search()
    # print(f"{monosearch.name}:  {monoresult}\n")

    indysearch = IndividualStatsSearch(2022, "11", "8", "ou", "scIzor")
    indyresult = indysearch.find_individual_usage()
    print(f"{indysearch.name}:  {indyresult}\n")

    # print(indyresult['rank'])

    # print(result.keys())



import pytest
import datetime


def test_mono_search_year(mono_stats_search):
    assert mono_stats_search.year == "2023"


def test_mono_search_month(mono_stats_search):
    assert mono_stats_search.month == "09"


def test_mono_search_typing(mono_stats_search):
    assert mono_stats_search.typing == "normal"


def test_mono_search_gen(mono_stats_search):
    assert mono_stats_search.gen == 9

def test_mono_search_url(mono_stats_search):
    assert mono_stats_search.base == "https://www.smogon.com/stats/2023-09/monotype/gen9monotype-mononormal-1500.txt"

def test_is_monotype(mono_stats_search):
    assert mono_stats_search.isMonotype == True

def test_mono_search_validation_object(mono_stats_search):
    assert mono_stats_search._create_validation_object() == {
        "year": "2023",
        "month": "09",
        "date": datetime.datetime(2023, 9, 1, 0, 0),
        "gen": 9,
        "base": "https://www.smogon.com/stats/2023-09/monotype/gen9monotype-mononormal-1500.txt",
        "ending": "gen9monotype-mononormal-1500.txt",
        "isMonotype": True,
        "typing": "normal",
    }



import sys
import os

# import smog_usage_stats.Search as Search
# import smog_usage_stats.Validation as Validation
# import smog_usage_stats.UsageStatsLookup as UsageStatsLookup
# import smog_usage_stats.IndividualLookup as IndividualLookup

# from smog_usage_stats.Search import Search
# from smog_usage_stats.UsageStatsLookup import StatsSearch, BaseStatsSearch, MonotypeStatsSearch
# from smog_usage_stats.IndividualLookup import ChaosSearch, BaseChaosSearch, MonotypeChaosSearch, IndividualStatsSearch
# from smog_usage_stats.Validation import Validations
# from smog_usage_stats.Update import Updater
# from smog_usage_stats.SQLInterface import SQLInterface

__version__ = "1.0.8"
__author__ = "Alan Nardo"

# Get the parent directory
parent_dir = os.path.dirname(os.path.realpath(__file__))

# Add the parent directory to sys.path
sys.path.append(parent_dir)



from typing import Literal
from os.path import dirname as up
import os
import shutil
from datetime import datetime
import csv


# "Mega"-parent searching class. Everything contained within this class will be available,
# but not necessarily used by child classes.
class Search:
    """Parent searcher class.

    ::params:
    >year (str or int): a string or integer year
    >month (str): two-digit string month
    >gen: (str or int): an integer for whichever Pokemon generation is being queried."""

    def __init__(
        self,
        year: str | int,
        month: Literal["Must be a two digit month string eg: '01' for January"],
        gen: str | int,
    ) -> None:
        self.year = year
        self.month = month
        self.date = datetime.strptime(("-".join([str(year), month])), "%Y-%m")
        self.gen = gen
        self.base = "https://www.smogon.com/stats/"
        self.isMonotype = False

    @property
    def year(self) -> str | int:
        return self._year

    @property
    def month(self) -> str:
        return self._month

    @property
    def gen(self) -> str:
        return self._gen

    @gen.setter
    def gen(self, value):
        self._gen = int(value)

    @year.setter
    def year(self, value):
        self._year = value

    @month.setter
    def month(self, value):
        self._month = value

    def _save_output(
        self,
        data: list[list],
        ending: str,
        pathname: str | os.PathLike = None,
    ) -> None:
        if pathname:
            storage_dir = os.path.join(
                f"{self._locate_base_data_directory()}\\", pathname
            )
        else:
            storage_dir = os.path.join(f"{self._locate_base_data_directory()}\\cache")

        if not os.path.exists((storage_dir)):
            os.mkdirs(storage_dir)
        # make the document
        filepath = os.path.join(storage_dir, f"{self.year}-{self.month}_{ending}.csv")
        with open(
            filepath,
            "w",
            newline="",
        ) as file:
            csvWriter = csv.writer(file, delimiter=",")
            csvWriter.writerows(data)

        if os.stat(filepath).st_size < 100:
            os.remove(filepath)
            print(f"File {filepath} deleted as it contained no data.")

    def search_and_save(self, pathname: str | os.PathLike = None) -> None:
        """
        Searches the smogon stats repo and saves files to system.
        """
        data = self.search()
        if self.ending:
            ending = self.ending.split("-")[0]
            self._save_output(data, ending, pathname=pathname)
        else:
            print("No data was saved.")
            return None

    def _create_validation_object(self):
        this = vars(self)
        validation_object = {k.replace("_", ""): v for k, v in this.items()}
        return validation_object

    def _locate_base_data_directory(
        self
    ) -> os.PathLike:
        """"""
        base_dir = up(up(up(".")))
        # set up the cached dir, theres probably a better way to do this but for now it will suffice
        cache_dir = os.path.join(base_dir, "data")
        return cache_dir

    @staticmethod
    def clear_cache() -> None:
        """Clears cache files if there are any."""
        base_dir = up(up("."))
        # set up the cached dir, theres probably a better way to do this but for now it will suffice
        cache_dir = os.path.join(base_dir, "data\\cache")
        if os.path.exists(cache_dir):
            shutil.rmtree(cache_dir)



import pytest
import datetime


def test_basic_search_year(basic_stats_search):
    assert basic_stats_search.year == "2023"


def test_basic_search_month(basic_stats_search):
    assert basic_stats_search.month == "09"


def test_basic_search_tier(basic_stats_search):
    assert basic_stats_search.tier == "ou"


def test_basic_search_gen(basic_stats_search):
    assert basic_stats_search.gen == 9

def test_basic_search_url(basic_stats_search):
    assert basic_stats_search.base == "https://www.smogon.com/stats/2023-09/gen9ou-1500.txt"

def test_is_monotype(basic_stats_search):
    assert basic_stats_search.isMonotype == False

def test_basic_search_validation_object(basic_stats_search):
    assert basic_stats_search._create_validation_object() == {
        "year": "2023",
        "month": "09",
        "date": datetime.datetime(2023, 9, 1, 0, 0),
        "gen": 9,
        "base": "https://www.smogon.com/stats/2023-09/gen9ou-1500.txt",
        "isMonotype": False,
        "ending": "gen9ou-1500.txt",
        "tier": "ou",
    }
